import { Injectable, Inject } from "@angular/core"
import { Http, Headers } from "@angular/http"
import { Cookies } from "./cookies"

import "rxjs/add/operator/toPromise";

@Injectable()
export class RPCService {
    private server: string;
    private httpAuth: string;
    private cookies: Cookies;
    private uniqueId: number;
    private shouldManageSessionId: boolean;
    private context: Object;
    private headers: Headers;

    constructor(@Inject(Http) private http: Http) {
        this.cookies = new Cookies();
        this.uniqueId = 0;
        this.shouldManageSessionId = false;
        this.context = JSON.parse(localStorage.getItem("user_context")) || {
            "lang": "en_US"
        };
    }

    /**
     * 
     * @param url 
     * @param params 
     */
    private createRequest(url: string, params: any): any {
        this.uniqueId += 1;

        if (this.shouldManageSessionId) {
            params.session_id = this.cookies.getSessionId();
        }

        this.headers = new Headers({
            "Content-Type": "application/json",
            "X-Openerp-Session-Id": this.cookies.getSessionId(),
            "Authorization": "Basic " + btoa(`${this.httpAuth}`)
        });

        let data = {
            jsonrpc: "2.0",
            method: "call",
            params
        }

        return JSON.stringify(data);
    }

    /**
     * 
     * @param response 
     */
    private handleResponse(response: any): Promise<any> {
        response = response.json();

        if (!response.error) {
            return response.result;
        }

        let error = response.error;
        let trace = {
            title: "   ",
            message: "",
            fultrace: error
        }

        if (error.code === 200 && error.message === "Odoo Server Error" && error.data.name === "werkzeug.exceptions.NotFound") {
            trace.title = "page_not_found";
            trace.message = "HTTP Error";
        } else if ( (error.code === 100 && error.message === "Odoo Session Expired") || (error.code === 300 && error.message === "OpenERP WebClient Error" && error.data.debug.match("SessionExpiredException"))) {
            trace.title = "session_expired";
            this.cookies.deleteSessionId();
        } else if ( (error.message === "Odoo Server Error" && /FATAL:  database "(.+)" does not exist/.test(error.data.message))) {
            trace.title = "database_not_found";
            trace.message = error.data.message;
        } else if (error.data.name === "openerp.exceptions.AccessError") {
            trace.title = "AccessError";
            trace.message = error.data.message;
        } else {
            let split = ("" + error.data.fault_code).split("\n")[0].split(" -- ");
            if (split.length > 1) {
                error.type = split.shift();
                error.data.fault_code = error.data.fault_code.substr(error.type.length + 4);
            }

            if (error.code === 200 && error.type) {
                trace.title = error.type;
                trace.message = error.data.fault_code.replace(/\n/g, "<br />");
            } else {
                trace.title = error.message;
                trace.message = error.data.debug.replace(/\n/g, "<br />");
            }
        }

        return Promise.reject(trace);
    }

    /**
     * 
     * @param error 
     */
    private handleErrors(error: any): Promise<any> {
        return Promise.reject(error.message || error);
    }

    /**
     * 
     * @param configuration 
     */
    public init(configuration: any): void {
        this.server = configuration.server;
        this.httpAuth = configuration.http_auth || null;
    }

    /**
     * 
     * @param server 
     */
    public setServer(server: string): void {
        this.server = server;
    }

    /**
     * 
     * @param httpAuth 
     */
    public setHttpAuth(httpAuth: string): void {
        this.httpAuth = httpAuth;
    }

    /**
     * 
     * @param url 
     * @param params 
     */
    public sendRequest(url: string, params: Object): Promise<any> {
        let options = this.createRequest(url, params);

        return this.http.post(this.server + url, options, { headers: this.headers })
            .toPromise()
            .then(this.handleResponse)
            .catch(this.handleErrors);
    }

    /**
     * 
     */
    public getServerInfo(): Promise<any> {
        return this.sendRequest("/web/webclient/version_info", {});
    }

    /**
     * 
     */
    public getSessionInfo(): Promise<any> {
        return this.sendRequest("/web/session/get_session_info", {});
    }

    /**
     * 
     * @param db 
     * @param login 
     * @param password 
     */
    public login(db: string, login: string, password: string): any {
        let params = {
            db,
            login,
            password
        }

        return this.sendRequest("/web/session/authenticate", params).then(result => {
            if (!result.uid) {
                this.cookies.deleteSessionId();

                return Promise.reject({
                    title: "wrong_login",
                    message: "Username and password don't match",
                    fulltrace: result
                });
            }

            this.context = result.user_context;
            localStorage.setItem("user_context", JSON.stringify(this.context));

            this.cookies.setSessionId(result.session_id);

            return result;
        });    
    }

    /**
     * 
     * @param force 
     */
    public isLoggedIn(force: boolean = true): Promise<any> {
        if (!force) {
            return Promise.resolve(this.cookies.getSessionId.length > 0);
        }

        return this.getSessionInfo().then(result => {
            this.cookies.setSessionId(result.session_id);
            return !!result.uid;
        });
    }

    /**
     * 
     * @param force 
     */
    public logout(force: boolean = true): Promise<any> {
        this.cookies.deleteSessionId();

        if (force) {
            return this.getServerInfo().then(response => {
                if (!!response.db) {
                    return this.login(response.db, "", "");
                }
            })
        }

        return Promise.resolve();
    }

    /**
     * 
     */
    public databaseList(): Promise<any> {
        return this.sendRequest("/web/database/get_list", {});
    }

    /**
     * 
     * @param model 
     * @param domain 
     * @param fields 
     * @param limit 
     */
    public searchRead(model: string, domain: any, fields: any, limit?: number): Promise<any> {
        let params = {
            model,
            domain,
            fields,
            limit,
            context: this.context
        }

        return this.sendRequest("/web/dataset/search_read", params);
    }

    /**
     * 
     * @param context 
     */
    public updateContext(context: any): void {
        localStorage.setItem("user_context", JSON.stringify(context));
        let args = [[(<any>this.context).uid], context];
        this.call("res.users", "write", args, {})
            .then( () => this.context = context)
            .catch((err: any) => this.context = context);
    }

    /**
     * 
     */
    public getContext(): any {
        return this.context;
    }

    /**
     * 
     * @param model 
     * @param method 
     * @param args 
     * @param kwargs 
     */
    public call(model: string, method: string, args: any, kwargs?: any): Promise<any> {
        kwargs = kwargs || {};
        kwargs.context = kwargs.context || {};
        Object.assign(kwargs.context, this.context);

        let params = {
            model,
            method,
            args,
            kwargs
        }

        return this.sendRequest("/web/dataset/call_kw", params);
    }
}
